/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "../headers.h"

BoundingBoxManager::BoundingBoxManager()
{
	boxList.name = "Box List";
}

BoundingBoxManager::~BoundingBoxManager()
{
	clear();
}

void BoundingBoxManager::clear()
{
	boxList.clear();
}

BoundingBoxManager *BoundingBoxManager::getInstance()
{
	return &instance;
}

void BoundingBoxManager::addBox(Entity *entity)
{
	entity->updateBoundingBox();
	
	Reference *ref = new Reference();
	ref->object = entity;
	
	boxList.add(ref);
}

void BoundingBoxManager::removeBox(Entity *entity)
{
	entity->updateBoundingBox();
	
	for (Reference *ref = (Reference*)boxList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		if (ref->object == entity)
		{
			ref = (Reference*)ref->previous;
			boxList.remove(ref->next);
		}
	}
}

Entity *BoundingBoxManager::getFirst(Entity *entity)
{
	iterator = (Reference*)boxList.getFirstElement();
	
	if (iterator != NULL)
	{
		return (Entity*)iterator->object;
	}
	
	return NULL;
}

Entity *BoundingBoxManager::getNext()
{
	iterator = (Reference*)iterator->next;
	
	if (iterator != NULL)
	{
		return (Entity*)iterator->object;
	}
	
	return NULL;
}

bool BoundingBoxManager::isValidSightTarget(int entityType)
{
	switch (entityType)
	{
		case ET_BLOB:
		case ET_BIOMECH:
		case ET_DOOR:
		case ET_STRUCTURE:
		case ET_WEAK_WALL:
		case ET_BARRIER:
		case ET_LIFT:
		case ET_AUTOLIFT:
			return true;
			break;
				
		default:
			return false;
			break;
	}
	
	return false;
}

bool BoundingBoxManager::canSightTarget(Entity *observer, Entity *target)
{
	Entity *entity;
	
	for (Reference *ref = (Reference*)boxList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		entity = (Entity*)ref->object;
		
		if ((entity == observer) || (entity == target))
		{
			continue;
		}
		
		if (!isValidSightTarget(entity->entityType))
		{
			continue;
		}
		
		if (lineIntersectBox(entity, observer->position, target->position))
		{
			return false;
		}
	}
	
	return true;
}

bool BoundingBoxManager::canSightSolid(Entity *observer, Vector startPosition, Vector target)
{
	Entity *entity;
	
	for (Reference *ref = (Reference*)boxList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		entity = (Entity*)ref->object;
		
		if (entity == observer)
		{
			continue;
		}
		
		if (!isValidSightTarget(entity->entityType))
		{
			continue;
		}
		
		if (lineIntersectBox(entity, startPosition, target))
		{
			return true;
		}
	}
	
	return false;
}

bool BoundingBoxManager::lineIntersectBox(Entity *e, Vector start, Vector end)
{
	char inside = 1, quadrant[3];
	int i, whichPlane;
	double maxT[3], candidatePlane[3], origin[3], minB[3], maxB[3], dir[3], coord[3], percentage;
	Vector v;
	
	v = end;
	v -= start;
	v.normalize();
	
	dir[0] = v.x;
	dir[1] = v.y;
	dir[2] = v.z;
	
	origin[0] = start.x;
	origin[1] = start.y;
	origin[2] = start.z;
	
	minB[0] = e->position.x + e->boundingBox.mins.x;
	minB[1] = e->position.y + e->boundingBox.mins.y;
	minB[2] = e->position.z + e->boundingBox.mins.z;
	
	maxB[0] = e->position.x + e->boundingBox.maxs.x;
	maxB[1] = e->position.y + e->boundingBox.maxs.y;
	maxB[2] = e->position.z + e->boundingBox.maxs.z;
	
	/* Find candidate planes */
	
	for (i=0;i<3;i++)
	{
		if (origin[i] < minB[i])
		{
			quadrant[i] = 1;
			candidatePlane[i] = minB[i];
			inside = 0;
		}
		
		else if (origin[i] > maxB[i])
		{
			quadrant[i] = 0;
			candidatePlane[i] = maxB[i];
			inside = 0;
		}
		
		else
		{
			quadrant[i] = 2;
		}
	}

	/* Ray origin inside bounding box */
	
	if (inside == 1)
	{
		coord[0] = origin[0];
		coord[1] = origin[1];
		coord[2] = origin[2];
		
		return true;
	}

	/* Calculate T distances to candidate planes */
	
	for (i = 0; i < 3; i++)
	{
		if (quadrant[i] != 2 && dir[i] != 0)
		{
			maxT[i] = (candidatePlane[i] - origin[i]) / dir[i];
		}
		
		else
		{
			maxT[i] = -1.;
		}
	}

	/* Get largest of the maxT's for final choice of intersection */
	
	whichPlane = 0;
	
	for (i=1;i<3;i++)
	{
		if (maxT[whichPlane] < maxT[i])
		{
			whichPlane = i;
		}
	}

	/* Check final candidate actually inside box */
	
	if (maxT[whichPlane] < 0)
	{
		return false;
	}
	
	for (i=0;i<3;i++)
	{
		if (whichPlane != i)
		{
			coord[i] = origin[i] + maxT[whichPlane] * dir[i];
			
			if (coord[i] < minB[i] || coord[i] > maxB[i])
			{
				return false;
			}
		}
		
		else
		{
			coord[i] = candidatePlane[i];
		}
		
		/* ray hits box */
	}
	
	if (dir[0] != 0)
	{
		percentage = fabs(coord[0] - start.x) / fabs(end.x - start.x);
		
		if (percentage > 1)
		{
			return false;
		}
		
		return true;
	}
	
	if (dir[1] != 0)
	{
		percentage = fabs(coord[1] - start.y) / fabs(end.y - start.y);
		
		if (percentage > 1)
		{
			return false;
		}
		
		return true;
	}
	
	if (dir[2] != 0)
	{
		percentage = fabs(coord[2] - start.z) / fabs(end.z - start.z);
		
		if (percentage > 1)
		{
			return false;
		}
		
		return true;
	}
	
	return true;
}

void BoundingBoxManager::checkToOthers(Entity *entity)
{
	
	
	//entity->position += entity->velocity;
}

BoundingBoxManager BoundingBoxManager::instance;
