/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "doors.h"

void doorTouchedNormal(Entity *other);
void doorTouchedClosing(Entity *other);

void doorClose(Door *door)
{
	door->velocity = door->startPosition;
	door->velocity -= door->position;
	door->velocity.normalize();
	door->velocity = door->velocity * door->speed;
	door->destination = door->startPosition;
	door->state = DOOR_CLOSING;
	
	door->touch = &doorTouchedClosing;
	door->autoCloseTimer = 300;
	
	audio->playSound(SND_DOOR_MOVE_START, CH_EFFECT9);
}

void doorOpen(Door *door)
{
	door->velocity = door->endPosition;
	door->velocity -= door->position;
	door->velocity.normalize();
	door->velocity = door->velocity * door->speed;
	door->destination = door->endPosition;
	door->state = DOOR_OPENING;
	
	door->touch = &doorTouchedNormal;
	
	audio->playSound(SND_DOOR_MOVE_START, CH_EFFECT9);
}

void doorTouchedNormal(Entity *other)
{
	Door *door = (Door*)self;
	
	if (door->state == DOOR_LOCKED)
	{
		if (other == player)
		{
			game->setGamePlayMessage(200, localeManager->translate("HINT_DOOR_LOCKED"));
		}
		return;
	}
	
	if ((door->state != DOOR_CLOSED) && (door->state != DOOR_CLOSING))
	{
		return;
	}
	
	if ((other->entityType != ET_BLOB) && (other->entityType != ET_BIOMECH))
	{
		return;
	}
	
	doorOpen(door);
}

void doorTouchedClosing(Entity *other)
{
	if ((other->entityType != ET_BLOB) && (other->entityType != ET_BIOMECH))
	{
		return;
	}
	
	Door *door = (Door*)self;
	
	doorOpen(door);
	door->autoClose = true;
	return;
}

void doorThink()
{
	Door *door = (Door*)self;
	
	if ((door->state == DOOR_OPEN) && (door->autoClose))
	{
		door->autoCloseTimer -= engine->getTimeDifference(TD_LOGIC);
		
		if (door->autoCloseTimer <= 0)
		{			
			doorClose(door);
			return;
		}
	}
	
	if ((door->state != DOOR_OPENING) && (door->state != DOOR_CLOSING))
	{
		return;
	}
	
	if (Math::getDistance(door->position, door->destination) < engine->getTimeDifference(TD_LOGIC))
	{
		door->position = door->destination;
		door->velocity.set(0, 0, 0);
		
		if (door->position == door->endPosition)
		{
			door->state = DOOR_OPEN;
		}
		else
		{
			door->touch = &doorTouchedNormal;
			
			door->state = DOOR_CLOSED;
			
			if (door->initialLockCount > 0)
			{
				door->state = DOOR_LOCKED;
			}
		}
		
		audio->playSound(SND_DOOR_MOVE_END, CH_EFFECT9);
	}
}

void addDoor(Properties *props)
{
	int modelIndex;
	
	sscanf(props->getString("model", "*0"), "%*c%d", &modelIndex);
	
	MDLModel *mdlModel = modelManager->getMDLModel(props->getString("model", "NULL"));
	BSPModel *bspModel = &bsp->models[modelIndex];
	
	if ((mdlModel == NULL) || (bspModel == NULL) || (modelIndex == 0))
	{
		props->listProperties();
		graphics->showErrorAndExit("createBaseStructure() - Bad BSP Model '%s'", props->getString("model", "NULL"));
	}
	
	Door *door = entityManager->spawnDoor();
	
	door->state = engine->getValueOfDefine(props->getString("state", "DOOR_CLOSED"));
	
	if (door->lockCount > 0)
	{
		door->state = DOOR_LOCKED;
	}
	
	if (door->state == DOOR_LOCKED)
	{
		door->lockCount = max(1, door->lockCount);
		door->initialLockCount = max(1, door->initialLockCount);
	}
	
	door->model = mdlModel;
	
	Vector v;
	
	v.x = ((mdlModel->maxs.x) - (mdlModel->mins.x)) / 2;
	v.y = ((mdlModel->maxs.y) - (mdlModel->mins.y)) / 2;
	v.z = ((mdlModel->maxs.z) - (mdlModel->mins.z)) / 2;
	
	door->boundingBox.mins -= v;
	door->boundingBox.maxs += v;
	
	door->position = mdlModel->originalPosition;
	door->startPosition = door->position;
	
	// default opening position
	door->endPosition = door->startPosition;
	door->endPosition.z += (v.z * 1.75);
	
	door->action = &doorThink;
	door->touch = &doorTouchedNormal;
	
	int textureId = bsp->faces[bspModel->faceIndex].textureID;
	
	door->textureSpec = bsp->textureSpec[textureId];
	
	door->load(props);
}

void drawDoor()
{
	glDisable(GL_LIGHTING);
	
	Door *door = (Door*)self;
	
	enableMultiTexturing();
	
	for (int i = 0 ; i < door->model->numOfFaces ; i++)
	{
		glPushMatrix();
		{
			glTranslatef(door->position.x, door->position.y, door->position.z);
			
			glVertexPointer(3, GL_FLOAT, sizeof(Vector), &door->model->position[i][0]);
		
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glTexCoordPointer(2, GL_FLOAT, sizeof(Point), &door->model->textureCoord[i][0]);
		
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glTexCoordPointer(2, GL_FLOAT, sizeof(Point), &door->model->lightmapCoord[i][0]);
		
			glActiveTextureARB(GL_TEXTURE0_ARB);
			door->textureSpec->activate(engine->getTimeDifference(TD_ANIMATION));
		
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("Grey")->data);
		
			glActiveTextureARB(GL_TEXTURE0_ARB);
		
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonOffset(1.0, 1.0);
			glDrawElements(GL_TRIANGLES, door->model->numMeshVerts[i], GL_UNSIGNED_INT, &door->model->indices[i][0]);
			glDisable(GL_POLYGON_OFFSET_FILL);
		
			bsp->textureSpec[bsp->faces[i].textureID]->finish();
		}
		glPopMatrix();
	}
	
	finishMultiTexturing();
}

void doDoor()
{
	drawDoor();
		
	self->prepareToMove(engine->getTimeDifference(TD_LOGIC));
	
	self->move(engine->getTimeDifference(TD_LOGIC));
	
	self->action();
}
