/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "../headers.h"

EntityManager::EntityManager()
{
	blobList.setName("BlobList");
	enemyList.setName("EnemyList");
	itemList.setName("ItemList");
	weaponList.setName("WeaponList");
	decorationList.setName("DecorationList");
	bulletList.setName("BulletList");
	featureList.setName("FeatureList");
	elementList.setName("ElementList");
	Q3TargetTable.name = "TargetTable";

	uniqueId = 0;
}

EntityManager *EntityManager::getInstance()
{
	return &instance;
}

EntityManager::~EntityManager()
{
}

void EntityManager::destroy()
{
	clear();
}

Unit *EntityManager::spawnBlob(int type)
{
	uniqueId++;
	
	BlobDef *def = &blobDef[type];

	Unit *unit = new Unit(uniqueId);
	unit->entityType = ET_BLOB;
	unit->definition = def;
	unit->health = unit->maxHealth = def->hp;
	unit->shield = unit->maxShield = def->shield;
	unit->flags = def->flags;
	
	unit->boundingBox.mins = def->bbMins;
	unit->boundingBox.maxs = def->bbMaxs;
	unit->updateBoundingBox();
	
	unit->currentWeapon = &weaponDef[def->weapon];
	
	blobList.add(unit);

	return unit;
}

Unit *EntityManager::spawnEnemy(int type)
{
	uniqueId++;
	
	EnemyDef *def = &enemyDef[type];

	Unit *unit = new Unit(uniqueId);
	unit->entityType = ET_BIOMECH;
	unit->definition = def;
	unit->health = unit->maxHealth = Math::rrand(def->minHP, def->maxHP);
	unit->shield = unit->maxShield = def->shield;
	unit->flags = def->flags;
	
	unit->boundingBox.mins = def->bbMins;
	unit->boundingBox.maxs = def->bbMaxs;
	unit->updateBoundingBox();
	
	unit->currentWeapon = &weaponDef[def->weapon];
	
	enemyList.add(unit);
	
	return unit;
}

Item *EntityManager::spawnItem(int type)
{
	uniqueId++;
	
	ItemDef *def = &itemDef[type];

	Item *item = new Item(uniqueId);
	item->definition = def;
	item->itemPower = def->itemPower;
	item->health = 9999;
	
	item->boundingBox.mins = def->bbMins;
	item->boundingBox.maxs = def->bbMaxs;
	item->updateBoundingBox();
	
	Math::addBit(&item->flags, EF_BOUNCES);
	
	itemList.add(item);
	
	return item;
}

Weapon *EntityManager::spawnWeapon(int type, bool fullClip)
{
	uniqueId++;
	
	WeaponDef *def = &weaponDef[type];

	Weapon *weapon = new Weapon(uniqueId);
	weapon->health = 9999;
	weapon->definition = def;
	
	weapon->boundingBox.mins = def->bbMins;
	weapon->boundingBox.maxs = def->bbMaxs;
	weapon->updateBoundingBox();
	
	if (fullClip)
	{
		weapon->currentAmmo = ((WeaponDef*)weapon->definition)->clipSize;
	}
	
	weaponList.add(weapon);

	return weapon;
}

Decoration *EntityManager::spawnDecoration(int type)
{
	Decoration *decoration = new Decoration();
	decorationList.add(decoration);
	
	decoration->type = type;

	return decoration;
}

Bullet *EntityManager::spawnBullet()
{
	Bullet *bullet = new Bullet();
	bulletList.add(bullet);
	
	return bullet;
}

Feature *EntityManager::spawnFeature(int featureType)
{
	Feature *feature = new Feature();
	featureList.add(feature);
	
	feature->definition = &featureDef[featureType];
	feature->featureType = featureType;
	feature->health = 999;
	feature->thinkTime = 0;
	
	return feature;
}

Door *EntityManager::spawnDoor()
{
	Door *door = new Door();
	elementList.add(door);
	
	return door;
}

Switch *EntityManager::spawnSwitch()
{
	Switch *swt = new Switch();
	elementList.add(swt);
	
	return swt;
}

Structure *EntityManager::spawnStructure(int type)
{
	Structure *structure = new Structure();
	structure->entityType = type;
	elementList.add(structure);
	
	return structure;
}

Q3Target *EntityManager::spawnQ3Target(char *name)
{
	Q3Target *target = new Q3Target();
	Q3TargetTable.put(name, target);
	
	return target;
}

Unit *EntityManager::getOwnerById(unsigned int uniqueId)
{
	for (Unit *unit = (Unit*)enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if (unit->getUniqueId() == uniqueId)
		{
			return unit;
		}
	}
	
	for (Unit *unit = (Unit*)blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if (unit->getUniqueId() == uniqueId)
		{
			return unit;
		}
	}
	
	printf("WARNING - EntityManager::getOwnerById(%d) returning NULL!\n", uniqueId);
	
	return NULL;
}

Unit *EntityManager::getEnemyByName(char *name)
{
	for (Unit *unit = (Unit*)enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if (unit->name == name)
		{
			return unit;
		}
	}
	
	printf("WARNING - EntityManager::getEnemyByName(%d) returning NULL!\n", uniqueId);
	
	return NULL;
}

Item *EntityManager::getItemById(unsigned int uniqueId)
{
	for (Item *item = (Item*)itemList.getFirstElement() ; item != NULL ; item = (Item*)item->next)
	{
		if (item->getUniqueId() == uniqueId)
		{
			return item;
		}
	}
	
	printf("WARNING - EntityManager::getItemById(%d) returning NULL!\n", uniqueId);
	
	return NULL;
}

Weapon *EntityManager::getWeaponById(unsigned int uniqueId)
{
	for (Weapon *weapon = (Weapon*)weaponList.getFirstElement() ; weapon != NULL ; weapon = (Weapon*)weapon->next)
	{
		if (weapon->getUniqueId() == uniqueId)
		{
			return weapon;
		}
	}
	
	printf("WARNING - EntityManager::getWeaponById(%d) returning NULL!\n", uniqueId);
	
	return NULL;
}

int EntityManager::getEntityType(Properties *props)
{
	if (props->hasProperty("definitionType"))
	{
		return props->getInt("definitionType", -1);
	}
	
	return Engine::getInstance()->getValueOfDefine(props->getString("classname", "classname"));
}

void EntityManager::clear()
{
	blobList.clear();
	enemyList.clear();
	itemList.clear();
	weaponList.clear();
	decorationList.clear();
	featureList.clear();
	bulletList.clear();
	elementList.clear();
	Q3TargetTable.clear();
	
	uniqueId = 0;
}

void EntityManager::load(Properties *props)
{
	uniqueId = props->getInt("uniqueId", 0);
}

void EntityManager::save(FILE *fp)
{
	debug(("EntityManager::save()\n"));
	
	Unit *unit;
	Entity *ent;
	
	debug(("EntityManager::save() - Blobs\n"));
	for (ent = (Entity*)blobList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Enemies\n"));
	for (unit = (Unit*)enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		unit->save(fp);
		
		if (unit->currentItem != NULL)
		{
			unit->currentItem->save(fp);
		}
	}
	
	debug(("EntityManager::save() - Items\n"));
	for (ent = (Entity*)itemList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Weapons\n"));
	for (ent = (Entity*)weaponList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Bullets\n"));
	for (ent = (Entity*)bulletList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Features\n"));
	for (ent = (Entity*)featureList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Decoration\n"));
	for (ent = (Entity*)decorationList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	debug(("EntityManager::save() - Elements\n"));
	for (ent = (Entity*)elementList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		ent->save(fp);
	}
	
	// Do this last to keep the Entity Manager in sync upon a reload!
	Properties props;
	props.setName("EntityManager");
	props.setProperty("uniqueId", uniqueId);
	props.save(fp);
	
	debug(("EntityManager::save() - Done\n"));
}

EntityManager EntityManager::instance;
