/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "structures.h"

void moveStructureToStart(Structure *structure)
{
	structure->velocity = structure->startPosition;
	structure->velocity -= structure->endPosition;
	structure->velocity.normalize();
	structure->velocity = structure->velocity * structure->speed;
	structure->destination = structure->startPosition;
}

void moveStructureToEnd(Structure *structure)
{
	structure->velocity = structure->endPosition;
	structure->velocity -= structure->startPosition;
	structure->velocity.normalize();
	structure->velocity = structure->velocity * structure->speed;
	structure->destination = structure->endPosition;
}

void changeDoorState(char *name, int state)
{
	Door *door;
	
	bool printedMessage = false;
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if (entity->entityType != ET_DOOR)
		{
			continue;
		}
		
		if (entity->name != name)
		{
			continue;
		}
			
		door = (Door*)entity;
		
		if (door->state == DOOR_LOCKED)
		{
			if (state == DOOR_OPEN)
			{
				if (--door->lockCount > 0)
				{
					if (!printedMessage)
					{
						game->setInfoMessage(INFO_HINT, "%d more to go...", door->lockCount);
						printedMessage = true;
					}
					continue;
				}
				else
				{
					if (door->initialLockCount > 1)
					{
						cancelSwitches(name);
					}
				}
			}
			else if (door->lockCount > 0)
			{
				door->lockCount++;
				continue;
			}
		}
		
		door->autoClose = false;
		
		if (state == DOOR_OPEN)
		{
			doorOpen(door);
		}
		else
		{
			doorClose(door);
		}
	}
}

void toggleShield(char *name, bool state)
{
	Structure *structure;
	
	bool shutDown = false;
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if (entity->entityType != ET_BARRIER)
		{
			continue;
		}
		
		if ((entity->name != name) || (entity->name == ""))
		{
			continue;
		}
		
		structure = (Structure*)entity;
		
		structure->custom = !structure->custom;
		
		shutDown = true;
	}
	
	if (shutDown)
	{
		audio->playSound(SND_BARRIER_SHUT_DOWN, CH_ANY);
	}
}

void toggleStructurePosition(char *name)
{
	Structure *structure;
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if (entity->entityType != ET_STRUCTURE)
		{
			continue;
		}
		
		if (entity->name != name)
		{
			continue;
		}
		
		structure = (Structure*)entity;
		
		if (structure->position == structure->startPosition)
		{
			moveStructureToEnd(structure);
		}
		else if (structure->position == structure->endPosition)
		{
			moveStructureToStart(structure);
		}
	}
}

void handleStructureStateChange(char *name, bool active)
{
	if (active)
	{
		changeDoorState(name, DOOR_OPEN);
		toggleStructurePosition(name);
		toggleShield(name, true);
		toggleLiftPosition(name, active);
	}
	else
	{
		changeDoorState(name, DOOR_CLOSED);
		toggleStructurePosition(name);
		toggleShield(name, false);
		toggleLiftPosition(name, active);
	}
}

// ===================================== General ================================

void destroyableStructureDie()
{
	Vector v = self->position;
	v.x += Math::rrand((int)self->boundingBox.mins.x, (int)self->boundingBox.maxs.x);
	v.y += Math::rrand((int)self->boundingBox.mins.y, (int)self->boundingBox.maxs.y);
	v.z += Math::rrand((int)self->boundingBox.mins.z, (int)self->boundingBox.maxs.z);
	
	doExplosion(v, Math::rrand(2, 5), self);
	
	self->health -= 10;
	
	if (self->health <= -100)
	{
		Math::addBit(&self->flags, EF_DEAD);
		doExplosion(v, 10, self);
		return;
	}
	
	self->thinkTime = Math::rrand(10, 50);
}

void destroyableStructurePostDie()
{
	self->health = 0;
	
	self->thinkTime = 0;
	
	self->action = destroyableStructureDie;
	
	Math::addBit(&self->flags, EF_DYING);
	
	fireTriggers(self->getName(), TRIGGER_TYPE_ENTITY_KILLED);
}

// =================================== Windows =================================

void splitWindow(Triangle *parent, int level)
{
	level++;
	
	if (level >= 8)
	{
		addGlassShard(parent);
		return;
	}
	
	if ((level >= 5) && ((rand() % 3) == 0))
	{
		addGlassShard(parent);
		return;
	}
	
	int vertex = rand() % 3;
	
	Vector split;
	Triangle child1, child2;
	
	switch (vertex)
	{
		case 0:
			split.x = parent->point[1].x + parent->point[2].x;
			split.y = parent->point[1].y + parent->point[2].y;
			split.z = parent->point[1].z + parent->point[2].z;
			break;
		
		case 1:
			split.x = parent->point[0].x + parent->point[2].x;
			split.y = parent->point[0].y + parent->point[2].y;
			split.z = parent->point[0].z + parent->point[2].z;
			break;
		
		case 2:
			split.x = parent->point[1].x + parent->point[0].x;
			split.y = parent->point[1].y + parent->point[0].y;
			split.z = parent->point[1].z + parent->point[0].z;
			break;
	}
	
	split.x /= 2;
	split.y /= 2;
	split.z /= 2;
	
	switch (vertex)
	{
		case 0:
			child1.point[0] = parent->point[0];
			child1.point[1] = parent->point[1];
			child1.point[2] = split;
			
			child2.point[0] = parent->point[0];
			child2.point[1] = split;
			child2.point[2] = parent->point[2];
			break;
		
		case 1:
			child1.point[0] = parent->point[0];
			child1.point[1] = parent->point[1];
			child1.point[2] = split;
			
			child2.point[0] = split;
			child2.point[1] = parent->point[1];
			child2.point[2] = parent->point[2];
			break;
		
		case 2:
			child1.point[0] = parent->point[0];
			child1.point[1] = split;
			child1.point[2] = parent->point[2];
			
			child2.point[0] = split;
			child2.point[1] = parent->point[1];
			child2.point[2] = parent->point[2];
			break;
	}
	
	splitWindow(&child1, level);
	splitWindow(&child2, level);
}

void windowShatter()
{
	debug(("windowShatter()\n"));
	
	Structure *structure = (Structure*)self;
	
	Triangle t1, t2;
	
	bool useX = (max(structure->boundingBox.maxs.x, structure->boundingBox.maxs.y) == structure->boundingBox.maxs.x);
	
	// A
	t1.point[0] = structure->position;
	(useX) ? t1.point[0].x += structure->boundingBox.mins.x : t1.point[0].y += structure->boundingBox.mins.y;
	t1.point[0].z += structure->boundingBox.maxs.z;
	
	// B
	t1.point[1] = structure->position;
	(useX) ? t1.point[1].x += structure->boundingBox.maxs.x : t1.point[1].y += structure->boundingBox.maxs.y;
	t1.point[1].z += structure->boundingBox.maxs.z;
	
	// D
	t1.point[2] = structure->position;
	(useX) ? t1.point[2].x += structure->boundingBox.mins.x : t1.point[2].y += structure->boundingBox.mins.y;
	t1.point[2].z += structure->boundingBox.mins.z;
	
	
	// B
	t2.point[0] = structure->position;
	(useX) ? t2.point[0].x += structure->boundingBox.maxs.x : t2.point[0].y += structure->boundingBox.maxs.y;
	t2.point[0].z += structure->boundingBox.maxs.z;
	
	// C
	t2.point[1] = structure->position;
	(useX) ? t2.point[1].x += structure->boundingBox.maxs.x : t2.point[1].y += structure->boundingBox.maxs.y;
	t2.point[1].z += structure->boundingBox.mins.z;
	
	// D
	t2.point[2] = structure->position;
	(useX) ? t2.point[2].x += structure->boundingBox.mins.x : t2.point[2].y += structure->boundingBox.mins.y;
	t2.point[2].z += structure->boundingBox.mins.z;
	
	engine->startDebugTimer("Shatter Window");
	splitWindow(&t1, 0);
	splitWindow(&t2, 0);
	engine->endDebugTimer("Shatter Window");
	
	Math::addBit(&self->flags, EF_DEAD);
}

// ================== Weak Walls ==============================

void weakWallLookForPlayer()
{
	if (!(player->flags & EF_ONGROUND))
	{
		return;
	}
	
	if (player->health < 0)
	{
		return;
	}
	
	int size;
	
	size = max((int)self->boundingBox.maxs.x, (int)self->boundingBox.maxs.y);
	size = max(size, (int)self->boundingBox.maxs.x);
	size = max(size, (int)self->boundingBox.maxs.y);
	size += 12;
	
	if (Math::getDistance(player->position, self->position) <= size)
	{
		if (!game->hasItem("C4"))
		{
			game->setGamePlayMessage(localeManager->translate("HINT_NEED_C4"));
			return;
		}
		
		game->setGamePlayMessage(localeManager->translate("HINT_USE_C4"));
		
		if (game->isControl(CTRL_ACTION))
		{
			windowShatter();
			
			self->health = -100;
			
			for (int i = 0 ; i < 35 ; i++)
			{
				throwExplosionDebris(self->position);
			}
			
			game->removeItem("C4");
			
			fireTriggers(self->getName(), TRIGGER_TYPE_ENTITY_KILLED);
		}
	}
}

// ================== General Structure Think =================

void structureThink()
{
	if (Math::getDistance(self->position, self->destination) < engine->getTimeDifference(TD_LOGIC))
	{
		self->position = self->destination;
		self->velocity.set(0, 0, 0);
	}
}

Structure *createBaseStructure(Properties *props, int type)
{
	int modelIndex;
	
	sscanf(props->getString("model", "*0"), "%*c%d", &modelIndex);
	
	MDLModel *mdlModel = modelManager->getMDLModel(props->getString("model", "NULL"));
	BSPModel *bspModel = &bsp->models[modelIndex];
	
	if ((mdlModel == NULL) || (bspModel == NULL) || (modelIndex == 0))
	{
		props->listProperties();
		graphics->showErrorAndExit("createBaseStructure() - Bad BSP Model '%s'", props->getString("model", "NULL"));
	}
	
	Structure *structure = entityManager->spawnStructure(type);
	
	structure->model = mdlModel;
	
	Vector v;
	
	v.x = ((mdlModel->maxs.x) - (mdlModel->mins.x)) / 2;
	v.y = ((mdlModel->maxs.y) - (mdlModel->mins.y)) / 2;
	v.z = ((mdlModel->maxs.z) - (mdlModel->mins.z)) / 2;
	
	structure->boundingBox.mins -= v;
	structure->boundingBox.maxs += v;
	
	structure->position = mdlModel->originalPosition;
	structure->startPosition = structure->position;
	structure->endPosition = structure->position;
	
	int textureId = bsp->faces[bspModel->faceIndex].textureID;
	
	structure->textureSpec = bsp->textureSpec[textureId];
	
	return structure;
}

void addStructure(Properties *props)
{
	debug(("addStructure()\n"));
	
	Structure *structure = createBaseStructure(props, ET_STRUCTURE);
	
	structure->health = 1;
	
	Math::addBit(&structure->flags, EF_NOCLIP);
	
	if (props->getInt("canBeHurt", 0) == 1)
	{
		Math::removeBit(&structure->flags, EF_IMMORTAL);
	}
	
	structure->action = &structureThink;
	
	structure->load(props);
	
	if (!(structure->flags & EF_IMMORTAL))
	{
		structure->die = destroyableStructurePostDie;
	}
}

void addWindow(Properties *props)
{
	debug(("addWindow()\n"));
	
	Structure *window = createBaseStructure(props, ET_WINDOW);
	
	Math::removeBit(&window->flags, EF_IMMORTAL);
	window->health = Math::rrand(2, 5);
	
	window->load(props);
}

void addLift(Properties *props)
{
	debug(("addLift()\n"));
	
	Structure *lift = createBaseStructure(props, ET_LIFT);
	
	lift->action = &liftMainThink;
	lift->touch = &liftTouch;
	
	lift->destination = lift->position;
	lift->speed = 0.5;
	
	lift->load(props);
	
	lift->thinkTime = 0;
}

void addAutoLift(Properties *props)
{
	debug(("addAutoLift()\n"));
	
	Structure *lift = createBaseStructure(props, ET_AUTOLIFT);
	
	lift->custom = props->getInt("waitTime", 300);
	
	lift->action = &liftMainThink;
	lift->touch = &liftTouch;
	
	lift->destination = lift->position;
	lift->speed = 0.5;
	
	lift->load(props);
	
	lift->thinkTime = 0;
}

void addEnergyBarrier(Properties *props)
{
	Structure *barrier = createBaseStructure(props, ET_BARRIER);
	
	barrier->health = 250;
	barrier->custom = 1; // barrier is on
	
	barrier->load(props);
}

void addPushBlock(Properties *props)
{
	Structure *pushBlock = createBaseStructure(props, ET_PUSHBLOCK);
	
	pushBlock->name = "PushBlock";
	
	Math::removeBit(&pushBlock->flags, EF_NOCLIP);
	Math::removeBit(&pushBlock->flags, EF_NOWORLDCLIP);
	Math::removeBit(&pushBlock->flags, EF_WEIGHTLESS);
	Math::addBit(&pushBlock->flags, EF_PUSHABLE);
	
	pushBlock->load(props);
}

void addWeakWall(Properties *props)
{
	debug(("addWeakWall()\n"));
	
	Structure *wall = createBaseStructure(props, ET_WEAK_WALL);
	
	wall->health = RAND_MAX;
	wall->name = "WeakWall";
	
	wall->action = &weakWallLookForPlayer;
	wall->die = &destroyableStructureDie;
	
	wall->load(props);
	
	wall->thinkTime = 0;
}

void drawStructure()
{
	Structure *structure = (Structure*)self;
	
	glDisable(GL_LIGHTING);
	
	enableMultiTexturing();
	
	float fade = (structure->health / 500);
	
	for (int i = 0 ; i < structure->model->numOfFaces ; i++)
	{
		glPushMatrix();
		{
			glTranslatef(structure->position.x, structure->position.y, structure->position.z);
			
			glVertexPointer(3, GL_FLOAT, sizeof(Vector), &structure->model->position[i][0]);
		
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glTexCoordPointer(2, GL_FLOAT, sizeof(Point), &structure->model->textureCoord[i][0]);
		
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glTexCoordPointer(2, GL_FLOAT, sizeof(Point), &structure->model->lightmapCoord[i][0]);
		
			glActiveTextureARB(GL_TEXTURE0_ARB);
			structure->textureSpec->activate(engine->getTimeDifference(TD_ANIMATION));
		
			glActiveTextureARB(GL_TEXTURE1_ARB);
			if ((structure->entityType == ET_WINDOW) || (structure->entityType == ET_BARRIER))
			{
				glBindTexture(GL_TEXTURE_2D, bsp->whiteTexture);
			}
			else
			{
				glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("Grey")->data);
			}
		
			glActiveTextureARB(GL_TEXTURE0_ARB);
			
			if (structure->entityType == ET_BARRIER)
			{
				glColor4f(1.0, 1.0, 1.0, fade);
			}
		
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonOffset(1.0, 1.0);
			glDrawElements(GL_TRIANGLES, structure->model->numMeshVerts[i], GL_UNSIGNED_INT, &structure->model->indices[i][0]);
			glDisable(GL_POLYGON_OFFSET_FILL);
		
			bsp->textureSpec[bsp->faces[i].textureID]->finish();
		}
		glPopMatrix();
	}
	
	finishMultiTexturing();
}

void doStructure()
{
	drawStructure();
	
	Structure *structure = (Structure*)self;
	
	if ((structure->active) || (!(structure->flags & EF_WEIGHTLESS)))
	{
		structure->applyGravity(engine->getTimeDifference(TD_LOGIC));
		
		moveEntity(structure);
		
		if (structure->position.z < bsp->minCords.z)
		{
			printf("ERROR: Structure '%s' fell out of map\n", self->getName());
			exit(1);
		}
	}
	
	if (structure->flags & EF_PUSHABLE)
	{
		structure->velocity.x = structure->velocity.x * 0.85;
		structure->velocity.y = structure->velocity.y * 0.85;
		
		if ((structure->velocity.x > -EPSILON) && (structure->velocity.x < EPSILON))
		{
			structure->velocity.x = 0;
		}
	
		if ((structure->velocity.y > -EPSILON) && (structure->velocity.y < EPSILON))
		{
			structure->velocity.y = 0;
		}
		
		if (tracer->outputContent & (CONTENTS_WATER|CONTENTS_SLIME|CONTENTS_LAVA))
		{
			Math::limit(&structure->velocity.x, -2, 2);
			Math::limit(&structure->velocity.y, -2, 2);
			Math::limit(&structure->velocity.z, -2, 2);
		}
	}
	
	if (structure->thinkTime == 0)
	{
		if (structure->action != NULL)
		{
			structure->action();
		}
	}
}
