/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "switches.h"

void switchLookForPlayer();

bool allPressurePlatesDown(char *name)
{
	Switch *swt = (Switch*)self;
	
	if (strcmp(name, "") == 0)
	{
		printf("WARNING: allPressurePlatesDown() - Pressure Plates MUST have names!\n");
		return false;
	}
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if ((entity->entityType != ET_SWITCH) || (entity->name != name))
		{
			continue;
		}
		
		swt = (Switch*)entity;
		
		if (swt->switchType != SWT_PRESSURE)
		{
			continue;
		}
		
		if (!swt->active)
		{
			debug(("allPressurePlatesDown - '%s' - FALSE\n", swt->getName()));
			
			return false;
		}
	}
	
	debug(("allPressurePlatesDown - '%s' - TRUE\n", swt->getName()));
	
	return true;
}

void pressurePlateTouch(Entity *other)
{
	Switch *swt = (Switch*)self;
	
	if (!swt->active)
	{
		audio->playSound(SND_SWITCH_CLICK_3, CH_ANY);
		
		swt->active = true;
		
		bool enable = allPressurePlatesDown(swt->getName());
	
		handleStructureStateChange(swt->getName(), enable);
	}
	
	swt->thinkTime = 5;
}

void pressurePlateDeactivate()
{
	Switch *swt = (Switch*)self;
	
	if (swt->active)
	{
		swt->active = false;
		handleStructureStateChange(swt->getName(), false);
	}
}

void switchDoNothing()
{
	self->thinkTime = RAND_MAX;
}

void cancelSwitches(char *name)
{
	Switch *swt;
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if ((entity->entityType != ET_SWITCH) || (entity->name != name))
		{
			continue;
		}
		
		swt = (Switch*)entity;
		
		swt->switchType = SWT_STATIC;
		swt->action = switchDoNothing;
		
		debug(("cancelSwitches() - %s\n", name));
	}
}

void switchWait()
{
	self->action = &switchLookForPlayer;
	
	((Switch*)self)->active = false;
}

void switchDoTimer()
{
	Switch *swt = (Switch*)self;
	
	swt->active = false;
	
	handleStructureStateChange(swt->getName(), false);
	
	self->action = &switchLookForPlayer;
}

void handleSwitchUse(Switch *swt)
{
	if (swt->active)
	{
		return;
	}
	
	swt->active = true;
	
	audio->playSound(SND_SWITCH_BLEEP, CH_ANY);
			
	if (swt->name != "")
	{
		handleStructureStateChange(swt->getName(), swt->active);
	}
			
	fireTriggers(swt->getName(), TRIGGER_TYPE_ANY);
			
	switch (swt->switchType)
	{
		case SWT_NORMAL:
			swt->action = &switchDoNothing;
			break;
			
		case SWT_TOGGLE:
			swt->thinkTime = 100;
			swt->action = &switchWait;
			break;
					
		case SWT_TIMED:
			swt->thinkTime = swt->custom;
			swt->action = &switchDoTimer;
			break;
	}
}

void switchLookForPlayer()
{
	if ((player->health < 0) || (!(player->flags & EF_ONGROUND)))
	{
		return;
	}
	
	Switch *swt = (Switch*)self;
	
	if (Math::getDistance(player->position, swt->position) <= 10)
	{
		game->setGamePlayMessage(localeManager->translate("HINT_USE_SWITCH"));
		
		if (game->isControl(CTRL_ACTION))
		{
			game->resetControl(CTRL_ACTION);
			
			if (swt->requires != "")
			{
				if (!game->hasItem(swt->requires.getText()))
				{
					game->setInfoMessage(INFO_HINT, "%s required", swt->requires.getText());
					return;
				}
				else
				{
					game->removeItem(swt->requires.getText());
					game->setInfoMessage(INFO_HINT, "Used %s", swt->requires.getText());
					swt->requires = "";
				}
			}
			
			handleSwitchUse(swt);
		}
	}
}

void addSwitch(Properties *props)
{
	Switch *swt = entityManager->spawnSwitch();
	
	swt->load(props);
	
	swt->switchType = engine->getValueOfDefine(props->getString("switchType", "SWT_NORMAL"));
	
	switch (swt->switchType)
	{
		case SWT_NORMAL:
			if (swt->thinkTime == 0)
			{
				swt->action = &switchLookForPlayer;
			}
			else
			{
				swt->action = &switchWait;
			}
			break;
			
		case SWT_TIMED:
			if (swt->thinkTime == 0)
			{
				swt->action = &switchLookForPlayer;
			}
			else
			{
				swt->action = &switchDoTimer;
			}
			break;
			
		default:
			swt->action = &switchLookForPlayer;
			break;
	}
	
	debug(("addSwitch - %s\n", swt->position.toString()));
}

void addPressurePlate(Properties *props)
{
	Switch *swt = entityManager->spawnSwitch();
	
	swt->name = "PressurePlate";
	
	swt->load(props);
	
	swt->switchType = SWT_PRESSURE;
	
	swt->touch = &pressurePlateTouch;
	swt->action = &pressurePlateDeactivate;
	
	swt->boundingBox.mins.set(-16, -16, -1);
	swt->boundingBox.maxs.set(16, 16, 1);
	
	debug(("addSwitch - %s\n", swt->position.toString()));
}

void drawSwitch()
{
	static Vector position;
	static float width, length, height;
	
	Switch *swt = (Switch*)self;
	
	glDisable(GL_TEXTURE_2D);
	
	if (swt->switchType != SWT_PRESSURE)
	{
		(swt->active == 0) ? glColor3f(1.0, 0.0, 0.0) : glColor3f(0.0, 1.0, 0.0);
		
		width = swt->boundingBox.maxs.x + fabs(swt->boundingBox.mins.x);
		length = swt->boundingBox.maxs.y + fabs(swt->boundingBox.mins.y);
		height = swt->boundingBox.maxs.z + fabs(swt->boundingBox.mins.z);
		
		width /= 2;
		length /= 2;
		height /= 2;
		
		position = swt->position;
		position.z += height;
		
		drawBox(position, width * 2, length * 2, height * 2);
	}
	else
	{
		(swt->active) ? glColor3f(1.0, 1.0, 1.0) : glColor3f(0.5, 0.5, 0.5);
		
		drawBox(swt->position, 32, 32, 2);
	}
}

void doSwitch()
{
	bbManager->removeBox(self);
	
	if (bsp->checkToWorld(self))
	{
		bbManager->checkToOthers(self);
	}
	
	drawSwitch();
	
	if (self->thinkTime == 0)
	{
		self->action();
	}
	
	bbManager->addBox(self);
}
